/*
 * Copyright (c) 2019 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-Nordic-5-Clause
 */

/** @file
 *  @brief Demo
 */

#include <zephyr/types.h>
#include <stddef.h>
#include <inttypes.h>
#include <errno.h>
#include <zephyr.h>
#include <pm/pm.h>
#include <sys/printk.h>
#include <logging/log.h>

#include <drivers/gpio.h>
#include <hal/nrf_gpio.h>

#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/conn.h>
#include <bluetooth/uuid.h>
#include <bluetooth/gatt.h>
#include <bluetooth/gatt_dm.h>
#include <bluetooth/scan.h>
#include <bluetooth/controller.h>

#include <bluetooth/services/demo_service.h>
#include <dk_buttons_and_leds.h>

#include <events/sleep_event.h>
#include <settings/settings.h>

#include <demo_settings.h>

LOG_MODULE_REGISTER(main, LOG_LEVEL_DBG);

struct k_timer scan_timer;

void interrupts_init(void)
{

	/* Configure to generate PORT event (wakeup) */
	nrf_gpio_cfg_input(DT_GPIO_PIN(DT_NODELABEL(interrupt0), gpios),
			   NRF_GPIO_PIN_PULLDOWN);
	nrf_gpio_cfg_sense_set(DT_GPIO_PIN(DT_NODELABEL(interrupt0), gpios),
			       NRF_GPIO_PIN_SENSE_HIGH);
}

void go_to_sleep()
{       
        printk("Going to system off...\n");

        interrupts_init();
        pm_power_state_force((struct pm_state_info){PM_STATE_SOFT_OFF, 0, 0});

        LOG_ERR("System off failed");
}



void scan_timer_expired(struct k_timer *timer_id)
{
      struct sleep_event *event = new_sleep_event();
      
      printk("Scan timer expired\n");   
      
      EVENT_SUBMIT(event);
}

K_TIMER_DEFINE(scan_timer, scan_timer_expired, NULL);

static uint8_t count = 0;

int start_scanning()
{
      int err = 0;
      
      // Start scanning

      err = bt_scan_start(BT_SCAN_TYPE_SCAN_PASSIVE);
      if (err) {
              LOG_ERR("Scanning failed to start (err %d)\n", err);
              go_to_sleep();
      }

      printk("Scanning successfully started\n");      
     
      k_timer_start(&scan_timer, K_MSEC(5000), K_NO_WAIT);

      return err;
}

static struct bt_conn *default_conn;

static void scan_filter_match(struct bt_scan_device_info *device_info,
			      struct bt_scan_filter_match *filter_match,
			      bool connectable)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(device_info->recv_info->addr, addr, sizeof(addr));

	printk("Filters matched. Address: %s connectable: %s\n",
		addr, connectable ? "yes" : "no");
}

static void scan_connecting_error(struct bt_scan_device_info *device_info)
{
	printk("Connecting failed\n");
}

static void scan_connecting(struct bt_scan_device_info *device_info,
			    struct bt_conn *conn)
{
	default_conn = bt_conn_ref(conn);

        k_timer_stop(&scan_timer);
}

static void scan_filter_no_match(struct bt_scan_device_info *device_info,
				 bool connectable)
{
	int err;
	struct bt_conn *conn;
	char addr[BT_ADDR_LE_STR_LEN];

	if (device_info->recv_info->adv_type == BT_GAP_ADV_TYPE_ADV_DIRECT_IND) {
		bt_addr_le_to_str(device_info->recv_info->addr, addr,
				  sizeof(addr));
		printk("Direct advertising received from %s\n", addr);
		k_timer_stop(&scan_timer);
                bt_scan_stop();

		err = bt_conn_le_create(device_info->recv_info->addr,
					BT_CONN_LE_CREATE_CONN,
					device_info->conn_param, &conn);

		if (!err) {
                        default_conn = bt_conn_ref(conn);
			bt_conn_unref(conn);
		}
	}
}

BT_SCAN_CB_INIT(scan_cb, scan_filter_match, scan_filter_no_match,
		scan_connecting_error, scan_connecting);

static void connected(struct bt_conn *conn, uint8_t conn_err)
{
	char addr[BT_ADDR_LE_STR_LEN];
        
        bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	if (conn_err) {
		printk("Failed to connect to %s (%u)\n", addr, conn_err);
		if (conn == default_conn) {
			bt_conn_unref(default_conn);
			default_conn = NULL;

			start_scanning();
		}

		return;
	}

	printk("Connected: %s\n", addr);

}

static void disconnected(struct bt_conn *conn, uint8_t reason)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	printk("Disconnected: %s (reason %u)\n", addr, reason);

	if (default_conn != conn) {
		return;
	}

	bt_conn_unref(default_conn);
	default_conn = NULL;

        go_to_sleep();
}

static struct bt_conn_cb conn_callbacks = {
	.connected = connected,
	.disconnected = disconnected
};

static void scan_init(void)
{
	int err;

        struct bt_le_scan_param scan_params = {
                .type = BT_LE_SCAN_TYPE_PASSIVE,
                .options = BT_LE_SCAN_OPT_FILTER_DUPLICATE,
                .interval = BT_GAP_SCAN_FAST_INTERVAL,
                .window = BT_GAP_SCAN_FAST_WINDOW,
                .timeout = 0, // Not supported
                .interval_coded = 0,
                .window_coded = 0

        };

	struct bt_scan_init_param scan_init = {
		.connect_if_match = 1,
		.scan_param = &scan_params,
		.conn_param = BT_LE_CONN_PARAM_DEFAULT
	};

        struct bt_scan_short_name short_name = {
                .name = "Demo",
                .min_len = 4
        };	        

        bt_scan_init(&scan_init);            

	bt_scan_cb_register(&scan_cb);

	err = bt_scan_filter_add(BT_SCAN_FILTER_TYPE_SHORT_NAME,  &short_name);
	if (err) {
		printk("Scanning filters cannot be set (err %d)\n", err);

		return;
	}

	err = bt_scan_filter_enable(BT_SCAN_SHORT_NAME_FILTER, false);
	if (err) {
		printk("Filters cannot be turned on (err %d)\n", err);
	}
}

static void *app_count_cb(void)
{	
        count++;
        return &count;
}

static struct bt_ds_cb ds_callbacks = {
	.count_cb   = app_count_cb,
};

static bool event_handler(const struct event_header *eh)
{

  if (is_sleep_event(eh)) {
      
      go_to_sleep();

  }

  return true;
}

EVENT_LISTENER(main, event_handler);
EVENT_SUBSCRIBE(main, sleep_event);

/* Prevent deep sleep (system off) from being entered on long timeouts
 * or `K_FOREVER` due to the default residency policy.
 *
 * This has to be done before anything tries to sleep, which means
 * before the threading system starts up between PRE_KERNEL_2 and
 * POST_KERNEL.  Do it at the start of PRE_KERNEL_2.
 */
static int disable_ds_1(const struct device *dev)
{
	ARG_UNUSED(dev);

	pm_constraint_set(PM_STATE_SOFT_OFF);
	return 0;
}

SYS_INIT(disable_ds_1, PRE_KERNEL_2, 0);

void main(void)
{
	int err = 0;

	printk("Starting Demo\n");
        
        event_manager_init();

        bt_ctlr_set_public_addr(CENTRAL_BLE_ADDR_PUBLIC);

	err = bt_enable(NULL);
	if (err) {
		printk("Bluetooth init failed (err %d)\n", err);
		return;
	}

	printk("Bluetooth initialized\n");

        err = bt_ds_init(&ds_callbacks);
	if (err) {
		printk("Failed to init DS (err:%d)\n", err);
		return;
	}

	scan_init();
	bt_conn_cb_register(&conn_callbacks);

        interrupts_init();

        start_scanning();

        while(1);          

}
